#include <string.h>
#include <ctype.h>
#include "txtstr.h"

static const int MIN_EXTENT = 16;
static const int MAX_EXTENT = 1024;
static const int EXT_ALIGNMENT = 16;

TextString::TextString():
    capacity(0),
    len(0),
    str(0)
{
}

TextString::TextString(const TextString& line):
    capacity(line.capacity),
    len(line.len),
    str(0)
{
    if (capacity > 0) {
        str = new char[capacity];
        memmove(str, line.str, len+1);
    }
}


TextString::TextString(const char* line):
    capacity(0),
    len(0),
    str(0)
{
    setString(line);
}

TextString::~TextString() {
    delete str;
    // printf("Destructor ~TextString: this = %p, str=%p\n", this, str);
}

void TextString::setString(const char* line, int l) {
    if (line == 0) {
        delete[] str; str = 0;
        len = 0; capacity = 0;
        return;
    }

    if (l < 0)
        l = 0;

    if (capacity <= l || str == 0) {
        // Allocate a new buffer for a string
        capacity = l+1;
        delete[] str;
        str = new char[capacity];
    }
    memmove(str, line, l);
    len = l;
    str[len] = 0;
}

void TextString::setString(const char* line) {
    if (line == 0) {
        delete[] str; str = 0;
        len = 0; capacity = 0;
    } else {
        setString(line, strlen(line));
    }
}

TextString& TextString::operator=(const char* line) {
    setString(line, strlen(line));
    return *this;
}

TextString& TextString::operator=(const TextString& line) {
    setString(line.str, line.len);
    return *this;
}

void TextString::append(const char* line) {
    int l;
    if (line == 0 || (l = strlen(line)) == 0)
        return;
    ensureCapacity(len + l + 1);
    strcpy(str + len, line); // Append a line
    len += l;
}

void TextString::append(int c) {
    ensureCapacity(len + 2);
    str[len] = (char) c; str[len+1] = 0; // Append a character
    ++len;
}

void TextString::ensureCapacity(int n) {
    if (n > capacity) {
        int optimalExtent = n / 8;
        if (optimalExtent < MIN_EXTENT) {
            optimalExtent = MIN_EXTENT;
        } else if (optimalExtent > MAX_EXTENT) {
            optimalExtent = MAX_EXTENT;
        }
        capacity += optimalExtent;
        if (capacity < n)
            capacity = n;

        // Make capacity to be a multiple of EXT_ALIGNMENT
        if ((capacity % EXT_ALIGNMENT) != 0)
            capacity += (EXT_ALIGNMENT - capacity % EXT_ALIGNMENT);

        char* tmp = new char[capacity];
        if (str != 0) {
            memmove(tmp, str, len+1);
            delete[] str;
        }
        str = tmp;
    }
}

TextString& TextString::operator+=(const char* line) {
    append(line);
    return *this;
}

TextString& TextString::operator+=(int character) {
    append(character);
    return *this;
}

TextString TextString::operator+(const char* line) const {
    TextString t(*this);
    t.append(line);
    return t;
}

TextString TextString::operator+(int character) const {
    TextString t(*this);
    t.append(character);
    return t;
}

void TextString::insert(int pos, const char* line) {
    int l;
    if (line == 0 || (l = strlen(line)) == 0)
        return;
    ensureCapacity(len + l + 1);
    if (pos < 0)
        pos = 0;
    if (pos > len)
        pos = len;
    if (pos < len)
        memmove(str + pos + l, str + pos, len - pos);
    memmove(str + pos, line, l);
    len += l;
    str[len] = 0;
}

void TextString::insert(int pos, int c) {
    ensureCapacity(len + 2);
    if (pos < 0)
        pos = 0;
    if (pos > len)
        pos = len;
    if (pos < len)
        memmove(str + pos + 1, str + pos, len - pos);
    str[pos] = (char) c;
    ++len;
    str[len] = 0;
}

void TextString::removeAt(int pos) {
    if (pos >= len || pos < 0)
        return;
    if (pos < len-1)
        memmove(str + pos, str + pos + 1, (len-1) - pos);
    --len;
    str[len] = 0;
    truncate(len);
}

void TextString::setSize(int n) {
    if (n < 0)
        n = 0;
    ensureCapacity(n+1);
    len = n;
    str[len] = 0;
}

void TextString::truncate(int pos) {
    if (pos < len) {
        str[pos] = 0;
        len = pos;
    }
    if (capacity > len + 1 + EXT_ALIGNMENT) {
        // Release unused space
        int newCapacity = len + 1;
        if ((newCapacity % EXT_ALIGNMENT) != 0)
            newCapacity += (EXT_ALIGNMENT - newCapacity % EXT_ALIGNMENT);
        char* tmp = new char[newCapacity];
        memmove(tmp, str, len + 1);
        capacity = newCapacity;
        delete[] str;
        str = tmp;
    }
}

void TextString::trim() { // Remove white space at the end of line
    if (len == 0)
        return;
    int pos = len;
    while (pos > 0 && isspace(str[pos-1]))
        --pos;
    if (pos < len)
        truncate(pos);
}
